#  _jkmn.py v0.1 
#  Stefano Polla (polla@lorentz.leidenuniv.nl)
#  Code developed for AQA course tutorial and for personal research use
#  The code is not meant to be definitive and should be taken "as is".
#  Unit tests are not yet implemented


"""
    Module implementing the JKMN transform on fermionic operators
    (and, in the current version, additional utility methods).
    JKMN stands for Jiang, Kalev, Mruczkiewicz, Neven: the authors of
    arXiv:1910.10746 in which the ternary-tree transform is proposed.
"""

# TODO:
#  - Generic utilities -> move elsewhere
#  - Majorana operator utilities -> integrate in MajoranaOperator
#  - Conversions -> use openfermion functions when implemented
#  - jkmn -> integrate in openfermion.transforms


from openfermion import transforms, ops, utils
from openfermion import MolecularData
import numpy as np


# Generic utilities

def inline(symbolic_operator):
    ''' Get an inline string for a SymbolicOperator '''
    return str(symbolic_operator).replace('\n', ' ').replace('+ -', '- ')


# MajoranaOperator utilities

def compress_majorana(majorana_op, abs_tol=1e-08):
    ''' homologue of FermionOperator.compress(self, abs_tol) '''
    return ops.MajoranaOperator.from_dict({k: v for k, v in majorana_op.terms.items() if abs(v)>abs_tol })

def many_body_order_majorana(majorana_op):
    ''' homologue of FermionOperator.many_body_order(self) '''
    return max(len(k) for k in majorana_op.terms.keys())


# FermionOperator <-> MajoranaOperator conversions

def fermion_to_majorana(fermion_op):
    '''
    transforms a FermionOperator into the corresponding MajoranaOperator according
    to the following convention:
    
    c^(j)   -> [g(2j) + i g(2j+1)] / 2
    c(j)    -> [g(2j) - i g(2j+1)] / 2
    g(2j)   ->   [c(j) + c^(j)] 
    g(2j+1) -> i [c(j) - c^(j)] 
    
    (c fermions, g majoranas, ^ dagger, j index)
    '''
    majorana_op = ops.MajoranaOperator()
    for fermion_str, coeff in fermion_op.terms.items():
        majorana_tmp = ops.MajoranaOperator((), coeff)
        for fermion in fermion_str:
            if fermion[1] == 1:
                majorana_tmp *= ( ops.MajoranaOperator((2*fermion[0],), 0.5 ) + 
                             ops.MajoranaOperator((2*fermion[0]+1,), 0.5j) )
            else:
                majorana_tmp *= ( ops.MajoranaOperator((2*fermion[0],), 0.5) +
                             ops.MajoranaOperator((2*fermion[0]+1,), -0.5j) )
        majorana_op += majorana_tmp
        majorana_op = compress_majorana(majorana_op, abs_tol=0) # remove zero-terms (not automatic)
    return majorana_op

def majorana_to_fermion(majorana_op):
    '''
    transforms a FermionOperator into the corresponding MajoranaOperator according
    to the following convention:
    
    c^(j)   -> [g(2j) + i g(2j+1)] / 2
    c(j)    -> [g(2j) - i g(2j+1)] / 2
    g(2j)   ->   [c(j) + c^(j)] 
    g(2j+1) -> i [c(j) - c^(j)] 
    
    (c fermions, g majoranas, ^ dagger, j index)
    '''
    fermion_op = ops.FermionOperator()
    for majorana_str, coeff in majorana_op.terms.items():
        fermion_tmp = ops.FermionOperator((), coeff)
        for majorana in majorana_str:
            if (majorana % 2) == 0:
                fermion_tmp *= ( ops.FermionOperator(f'{majorana//2}', 1 ) + 
                                 ops.FermionOperator(f'{majorana//2}^', 1) )
            else:
                fermion_tmp *= ( ops.FermionOperator(f'{majorana//2}', 1j ) + 
                                 ops.FermionOperator(f'{majorana//2}^', -1j) )
                
        fermion_op += fermion_tmp
        fermion_op = utils.normal_ordered(fermion_op)
    return fermion_op


# JKMN transform

def _jkmn_single_majorana(majorana_idx, n_majorana_modes):
    ''' 
    internal function. 
    Implements the JKMN tree to convert a single majorana to a Pauli string.
    '''
    assert n_majorana_modes%2==0, 'number of majoranas must be even'
    assert 0<=majorana_idx<n_majorana_modes, 'index of the mode out of bounds'
    n_qubits = int(np.ceil(n_majorana_modes/2))
    h = int(np.log(2*n_qubits+1)/np.log(3))
    sigma = ('X', 'Y', 'Z')

    pauli_str = ''
    qubit_idx = 0
    old_branch = 0
    if majorana_idx < 3/2*(n_majorana_modes - (3**h-1)):
        heff = h+1
        tmp_idx = majorana_idx 
    else:
        heff = h
        tmp_idx = majorana_idx - (n_majorana_modes - (3**h-1))
    for level in range(heff):
        branch = int(tmp_idx // 3**(heff-1-level))
        pauli_str += sigma[branch] + str(qubit_idx) + ' '
        qubit_idx = (3**(level+1) - 1)//2 + 3*old_branch + branch
        tmp_idx = tmp_idx % int(3**(heff-1-level))
        old_branch = branch
    return pauli_str[:-1]

def jkmn(fermion_op, n_fermionic_modes):
    '''
        JKMN transform from arXiv:1910.10746
        Requires the specification of the number of fermionic modes.
        Returns a QubitOperator.
    '''
    n_majorana_modes = 2 * n_fermionic_modes
    majorana_op = fermion_to_majorana(fermion_op)
    qubit_op = ops.QubitOperator()
    
    for majorana_str, coeff in majorana_op.terms.items():
        qubit_tmp = ops.QubitOperator('', coeff)
        for majorana_idx in majorana_str:
            qubit_tmp *= ops.QubitOperator(_jkmn_single_majorana(majorana_idx, 
                                                                 n_majorana_modes))
        qubit_op += qubit_tmp
    return qubit_op